// The source code in this file was authored by
// individuals and organizations as listed below.
//
// Copyright (c) 2008  David Dindorp
//
// This source code is made available for use under
// the GPLv2 or GPLv3 license (at your discretion).
// See the file "COPYING" for details.  Contact the Free
// Software Foundation organization for more information.
// Some MySQL-specific exceptions apply, see the file "EXCEPTIONS".
// Do not remove or reduce this copyright notice.


using System;
using System.Collections.Generic;
using System.Threading;


/// <summary>
/// Drivers that wish to participate in pooling must implement this
/// interface.
/// </summary>
public interface IPoolableDriver {
	int ServerConnectionId { get; }
	/// <summary>
	/// Must indicate if the server supports flushing
	/// the connection state via Flush().
	/// </summary>
	bool FlushSupported { get; }
	/// <summary>
	/// Open connection to server.  May throw an exception.
	/// </summary>
	void Open();
	/// <summary>
	/// Close connection to server.  May throw an exception.
	/// </summary>
	void Close();
	/// <summary>
	/// Implement code that performs a wire protocol ping,
	/// and throws an exception when an error occurs.
	/// </summary>
	void Ping();
	/// <summary>
	/// Implement code that executes a server connection state flush,
	/// and throws an exception when an error occurs.
	/// </summary>
	void Flush();
	/// <summary>
	/// Call this event handler every time a query succeeds in order
	/// to clear the activity counter and prevent premture pings.
	/// </summary>
	event EventHandler QuerySuccess;
}


/// <summary>
/// Drivers that wish to participate in pooling must implement this
/// interface on top of their configuration (connection string) class.
/// </summary>
public interface IPoolConfiguration {
	/// <summary>
	/// GetHashCode() must return equality for
	/// connection strings that are equivalent.
	/// </summary>
	int GetHashCode();
	/// <summary>
	/// Equals must return equality for
	/// connection strings that are equivalent.
	/// </summary>
	bool Equals(object obj);
	int minPoolSize { get; }
	int maxPoolSize { get; }
	int pingInterval { get; }
	bool forceReuse { get; }
	int dormantPoolRetire { get; }
	int dormantDriverRetire { get; }
	string poolName { get; }
}


/// <summary>
/// Boilerplate event handler.
/// </summary>
public delegate void EventHandler();


/// <summary>
/// Singleton used by Connection objects when
/// they need to attach themselves to a pool.
///
/// Also used by Pool objects to tell the manager
/// when a pool wants to destroy itself.
/// </summary>
public class ConnectionPoolManager {
	public delegate IPoolableDriver DriverFactory(IPoolConfiguration config);
	// Connection pools are mapped to their config objects in this hash table.
	private Dictionary<IPoolConfiguration, ConnectionPool> Pools;
	// A lock is held when creating and destroying pools.
	private object poolLock;
	// Wrapper around a method that can create real pool driver objects.
	private DriverFactory factory;

	public ConnectionPoolManager(DriverFactory factory) {
		Pools = new Dictionary<IPoolConfiguration, ConnectionPool>();
		poolLock = new object();
		this.factory = factory;
	}

	/// <summary>
	/// Return new or existing pool.
	/// The configuration object must be comparable with
	/// other configuration objects, specifically it must
	/// override GetHashCode() and Equals() to return true
	/// for objects which has the same value.  The configuration
	/// object is used for mapping individual connections to
	/// each connection pool.
	/// </summary>
	public ConnectionPool GetPool(IPoolConfiguration config) {
		ConnectionPool pool = null;
		lock (poolLock) {
			if (! Pools.ContainsKey(config)) {
				// Performance counter eligible: one pool created.
				Pools[config] = new ConnectionPool(config, this);
			}
			pool = Pools[config];
		}
		return pool;
	}

	/// <summary>
	/// Remove an unused pool.
	/// </summary>
	internal void NukePool(IPoolConfiguration config) {
		lock (poolLock) {
			// Performance counter eligible: one pool destroyed.
			Pools.Remove(config);
		}
	}

	/// <summary>
	/// Used internally to create a pooled driver of the correct type,
	/// deriving from PooledDriverLink.
	/// </summary>
	internal PooledDriverLink CreateDriver(ConnectionPool pool, IPoolConfiguration config) {
		return new PooledDriverLink(pool, config, factory(config));
	}
}


/// <summary>
/// Maintains a set of driver objects corresponding to a given connection
/// string.
///
/// A driver is fetched from the pool by calling Catch().  After use,
/// the driver must be returned to the pool by calling Release().
///
/// When an in-use driver is released back to the pool, the server state
/// for the underlying server connection is first cleaned.  Then the object
/// is returned to the pool on the very top of the list of free drivers.
/// Thus the next Catch() call will most likely get a recently used driver.
///
/// After being released, if a driver sits in the pool for long enough
/// since the last executed statement, a ping to the server will occur over
/// the wire protocol.  If the ping is successful, the driver's
/// activity counter is reset.  If the ping is not successful, the driver
/// will either remove itself from the pool, or deliver an exception
/// describing the problem to the application (see below).
///
/// While a driver is pinging the server, it is not removed from the
/// pool of available drivers.  The application can call Catch() and
/// (in theory) hit a pinging driver.  When that happens, the app is
/// forced to wait for the ping to complete, which can take up to the
/// amount of time specified for the connection timeout.  If the ping
/// fails, an exception describing the problem is delivered to the app,
/// just as would happen if the driver was a newly constructed
/// instance trying to reach the database server.
///
/// The above will rarely happen in real life, for reasons noted below.
///
/// Applications with low levels of activity will rarely fetch a driver
/// that is performing a ping.  The default ping interval is 10 seconds,
/// and a ping takes on the order of milliseconds to perform, so it is
/// simply extremely unlikely.
///
/// Applications with a high level of activity causes the pool to rarely
/// if ever perform server pings, because drivers are fetched from the
/// pool so quickly after being released that the time since the last
/// driver activity rarely or never exceeds the ping interval.
///
/// To avoid excessive client and server load, the pool is not prepopulated
/// to the minimum pool size.  Instead, drivers are added as necessary.
/// Unused drivers are not purged beyond the minimum pool size, however.
///
/// An useful abstraction would be a "Min Pool Free" setting.  The pool
/// maintenance thread would at all times strive to hold at least as many
/// connections as specified hot and free for release.  Currently not
/// implemented, it would potentially improve application performance.
///
/// If an entire pool goes unused for a long time, it shuts itself down.
/// Tearing down long unused connection pools has two purposes.
///
/// It saves resources, which can be important if the connection strings
/// used to connect changes often.  This could be caused by a mechanism
/// such as automatically rotating passwords, or could just be a natural
/// high ratio of possible versus concurrently sustainable connection
/// pools.
///
/// It is also an important security measure, because it releases
/// pool memory containing login names and passwords for previously
/// used connections, so that the memory can be scrubbed by the memory
/// manager.  This is important if users authenticate in a manner where
/// the database password is derived from the user's login credentials,
/// such that keeping these in RAM forever even after the user has logged
/// out can be avoided.
/// </summary>
public class ConnectionPool {
	// Simple FIFO linked list of the drivers in this pool.
	protected volatile PooledDriverLink freeTop = null;
	protected object free = new Object();

	// Timer and thread used to destroy inactive pools.
	private EventTimer lastRetrieval = new EventTimer();
	private InterruptibleThread activityWatchdog;

	// Pool parameters.
	private IPoolConfiguration config;
	private ConnectionPoolManager manager;

	// Internal status.
	private volatile int poolSize = 0;
	private volatile int poolFree = 0;

	internal ConnectionPool(IPoolConfiguration config, ConnectionPoolManager manager) {
		this.config = config;
		this.manager = manager;
		StartActivityWatchdog();
	}

	/// <summary>
	/// Catch() is invoked to get a live, squiggly driver from the pool.
	/// </summary>
	///
	/// <exception cref="OverflowException">
	/// If the pool has already reached full capacity,
	/// an OverflowException is thrown.
	/// </exception>
	public PooledDriverLink Catch() {
		PooledDriverLink mine = null;
		lock (free) {
			// Grab the most recently released driver.
			mine = freeTop;
			if (freeTop != null) {
				freeTop = freeTop.next;
				poolFree--;
			}
		}
		if (mine == null) {
			// Create a new driver if none available.
			mine = manager.CreateDriver(this, config);
		} else {
			// Clear the now unused link to next free driver.
			mine.next = null;
		}
		mine.Bind();
		// Performance counter eligible: one driver caught.
		return mine;
	}

	/// <summary>
	/// Release() is invoked via the application to return
	/// a previously fetched driver to the pool.
	/// </summary>
	public void Release(ref PooledDriverLink dirty) {
		// This method can be called by external API
		// users, so handle bad values sensibly.
		if (dirty == null) return;
		// Reset driver state in the background.
		// The driver is responsible for re-adding
		// itself to the pool if successful.
		dirty.UnBind();
		// Performance counter eligible: one driver released.
		// Unassign reference to hint callers not to reuse.
		dirty = null;
	}

	/// <summary>
	/// New drivers internally call Register() when they're constructed.
	/// </summary>
	///
	/// <exception cref="OverflowException">
	/// If the pool has already reached full capacity,
	/// an OverflowException is thrown and ultimately
	/// propagated via Catch() to the application.
	/// </exception>
	internal void Register() {
		const string abortMsg = "The connection pool has reached it's maximum size.";
		lock (free) {
			if (poolSize >= config.maxPoolSize) throw new OverflowException(abortMsg);
			poolSize++;
		}
	}

	/// <summary>
	/// Aging drivers internally call UnRegister()
	/// when they wish to die of boredom.
	/// Free drivers which had a problem also calls
	/// UnRegister, with "important" set to true.
	/// </summary>
	internal bool UnRegister(PooledDriverLink driver, bool important) {
		PooledDriverLink current, prev;
		lock (free) {
			if (! important) {
				// Allow unregistering only if minimum pool size is satisfied.
				bool goingAway = lastRetrieval.HasTranspired(config.dormantPoolRetire);
				bool surplus = poolSize > config.minPoolSize;
				if (! (goingAway || surplus)) return false;
			}
			// Traverse the pool, looking for the driver.
			current = freeTop;
			prev = null;
			while (current != null && current != driver) {
				prev = current;
				current = current.next;
			}
			if (current == driver) {
				// Found! Unlink the driver.
				if (prev == null) freeTop = current.next;
				else prev.next = current.next;
				poolFree--;
			}
		}
		// Return false if the driver has already been caught
		// (end of list was reached).
		return current != null;
	}

	/// <summary>
	/// Drivers internally call BellyUp() when they croak.
	/// </summary>
	internal void BellyUp() {
		lock (free) {
			poolSize--;
		}
	}

	/// <summary>
	/// Released drivers internally call Swim() to place themselves
	/// at the top of the pool's free list immediately after they've
	/// finished cleaning themselves up.
	/// </summary>
	internal void Swim(PooledDriverLink fresh) {
		lock (free) {
			// Insert fresh entries at top of list.
			fresh.next = freeTop;
			freeTop = fresh;
			poolFree++;
		}
	}

	private void ActivityWatchdogThread(ManualResetEvent ready) {
		while (true) {
			try {
				ready.Set();
				// Wake up half a second after the pool life timer reaches
				// it's maximum, given that no connection has occurred
				// in the mean time.  Alternatively, wake up every half
				// second if waiting for in-use connections to be returned.
				int nextWake = lastRetrieval.NextEvent(config.dormantPoolRetire) * 1000 + 500;
				Thread.Sleep(nextWake);
				// When awoken, check if the pool has been used in the meantime.
				if (lastRetrieval.HasTranspired(config.dormantPoolRetire)) {
					// Make sure there are no unreleased drivers.
					lock (free) {
						// Are we good to go?
						if (poolSize == poolFree) {
							// Dereference pool from the pool manager.
							manager.NukePool(config);
							// Exit the maintenance thread.
							// After this, there should be no running threads
							// with references to the pool manager, except for
							// the associated drivers themselves, which should
							// be exiting shortly.  Garbage collection now
							// kicks in and takes all the objects away.
							break;
						}
					}
				}
			} catch (ThreadInterruptedException) {
				// For good measure - currently unused.
			}
		}
	}

	private void StartActivityWatchdog() {
		lastRetrieval.Reset();
		InterruptibleThread.WorkerCode watchdog = ActivityWatchdogThread;
		string name = "Pool watchdog " + config.poolName;
		activityWatchdog = new InterruptibleThread(name, 65536);
		activityWatchdog.Run(watchdog);
	}
}


public class PooledDriverLink {
	// Used to implement the linked list of free drivers.
	internal volatile PooledDriverLink next = null;

	// Pool to which this driver belongs, and driver reference.
	private ConnectionPool pool;
	private IPoolConfiguration config;
	private IPoolableDriver driver;

	// Used to avoid race condition when pinging/catching free drivers.
	private AutoResetEvent idle = new AutoResetEvent(true);

	// Used for bookkeeping by maintenance thread.
	private volatile bool isDirty = false;
	private EventTimer lastReleased = new EventTimer();
	private EventTimer lastActivity = new EventTimer();
	private volatile Exception pingProblem = null;
	private int delay;

	// Per-driver maintenance thread.
	private InterruptibleThread maintenanceThread;

	public IPoolableDriver Driver { get { return driver; } }

	internal PooledDriverLink(ConnectionPool pool, IPoolConfiguration config, IPoolableDriver driver) {
		this.pool = pool;
		this.config = config;
		this.driver = driver;
		// Connect to database server.
		Open();
		// Bind the query executed event handler to clear activity counter.
		driver.QuerySuccess += new EventHandler(ResetActivity);
		// Register with pool.
		pool.Register();
		// After registering, start the maintenance thread.
		StartMaintenanceThread();
	}

	/// <summary>
	/// For proper operation, must be called after each successful query.
	/// </summary>
	protected void ResetActivity() {
		lastActivity.Reset();
	}

	protected void Open() {
		driver.Open();
		lastActivity.Reset();
	}

	protected void SafeClose() {
		try {
			driver.Close();
		} catch { }
	}

	internal void Bind() {
		// When a connection is caught by the pool, this method
		// is called to grab the activity lock which ensures that
		// any background pings have finished executing.
		idle.WaitOne();
		if (pingProblem != null) throw pingProblem;
	}

	internal void UnBind() {
		// When a connection is released by the application, this
		// method is called to release the activity lock and start
		// the background cleaner.  But first, the release time is
		// recorded to prevent premature ejection from pool.
		lastReleased.Reset();
		idle.Set();
		// Set dirty flag and poke maintenance thread.
		isDirty = true;
		maintenanceThread.Interrupt();
	}

	private void MaintenanceThread(ManualResetEvent ready) {
		while (true) {
			try {
				ready.Set();
				// Sleep first.  When the driver is initially created,
				// allows the creator to grab the activity lock before we do.
				if (! isDirty) Thread.Sleep(delay);
				// Wait till noone is using the driver.
				idle.WaitOne();

				// Check if the driver has been lingering unused for too long.
				if ((! isDirty) && lastReleased.HasTranspired(config.dormantDriverRetire)) {
					// Make sure that minimum pool size is upheld.
					// First remove from pool, to prevent race condition with
					// an outsider that catches the driver from the pool.
					bool removed = pool.UnRegister(this, false);
					if (removed) {
						// Tear down connection, tell the pool we're gone,
						// and at last exit the maintenance thread.
						SafeClose();
						pool.BellyUp();
						break;
					} else {
						// Either the driver has already been caught and there
						// was really no need to go away after all, or the minimum
						// pool size would be exceeded.
					}
				}

				// Check if the driver is dirty and needs cleaning.
				if (isDirty) {
					if (config.forceReuse) {
						// Cleaning was explicitly disabled, just go back in.
						isDirty = false;
						pool.Swim(this);
					} else {
						try {
							if (driver.FlushSupported) {
								// Once the server supports this, we're ready.
								driver.Flush();
							} else {
								// Reconnect to clean state.
								SafeClose();
								Open();
							}
							// OK, good to go back in.
							isDirty = false;
							pool.Swim(this);
						} catch (Exception) {
							// Failed to clean state; instead of going
							// back into the pool, tell it we've croaked.
							SafeClose();
							pool.BellyUp();
							break;
						}
					}
				}

				// Check if it's time to make sure the connection's alive.
				if ((! isDirty) && lastActivity.HasTranspired(config.pingInterval)) {
					// Run a server ping.
					try {
						driver.Ping();
					} catch (Exception e) {
						// Store any problem for later consumption.
						pingProblem = e;
					}
					// In case of a problem, remove driver from the pool.
					if (pingProblem != null) {
						bool removed = pool.UnRegister(this, true);
						if (removed) {
							// Tear down connection as nicely as possible,
							// tell the pool we've quit, and at last exit
							// the maintenance thread.
							SafeClose();
							pool.BellyUp();
							break;
						} else {
							// The driver has already been caught.
							// Rest of the matter is handled in Bind().
						}
					}
				}

				// Clear the activity flag/lock.
				idle.Set();
			} catch (ThreadInterruptedException) {
				// Allow wake-ups when cleaning is needed.
				// If thread is interrupted while not sleeping,
				// it is postponed to the next sleep, and one
				// eventless extra round might be paced through.
			}
		}
	}

	private void StartMaintenanceThread() {
		// Calculate the maximum delay between loops.
		delay = Math.Min(config.pingInterval, config.dormantDriverRetire);
		if (config.pingInterval < 1) delay = config.dormantDriverRetire;
		if (config.dormantDriverRetire < 1) delay = config.pingInterval;
		if (delay < 1) return;
		// Activate half a second after the next event is supposed to transpire.
		delay = delay * 1000 + 500;
		InterruptibleThread.WorkerCode maintenance = MaintenanceThread;
		string name = "Driver maintenance " + driver.ServerConnectionId;
		maintenanceThread = new InterruptibleThread(name, 65536);
		maintenanceThread.Run(maintenance);
	}
}
