// The source code in this file was authored by
// individuals and organizations as listed below.
//
// Copyright (c) 2008  David Dindorp
//
// This source code is made available for use under
// the GPLv2 or GPLv3 license (at your discretion).
// See the file "COPYING" for details.  Contact the Free
// Software Foundation organization for more information.
// Some MySQL-specific exceptions apply, see the file "EXCEPTIONS".
// Do not remove or reduce this copyright notice.


using System;


/// <summary>
/// Uses a monotonically increasing value to implement an event timer.
/// This prevents a problem that could occur if DateTime.Now was
/// used, such as manual time changes or automatic time changes (DST)
/// causing time-dependent code to malfunction.
///
/// The timer must be reset at least once every 50 days, otherwise it
/// becomes imprecise on account of the underlying value wrapping.
///
/// The longest time you can wait for an event to transpire is also
/// just below 50 days, corresponding to uint.MaxValue milliseconds.
/// </summary>
public class EventTimer {
	private volatile uint value;

	public EventTimer() {
		Reset();
	}

	public void Reset() {
		value = (uint) Environment.TickCount;
	}

	public uint Elapsed {
		get {
			// Handles overflows.
			uint now = (uint)Environment.TickCount;
			uint elapsed = now < value ? uint.MaxValue - value + now : now - value;
			return elapsed / 1000;
		}
	}

	/// <summary>
	/// Has an event transpired?
	/// </summary>
	public bool HasTranspired(int seconds) {
		return Elapsed >= seconds;
	}

	/// <summary>
	/// How many seconds till a given event transpires?
	/// </summary>
	/// <returns>Returns 0 if the event has already come and gone.</returns>
	public int NextEvent(int seconds) {
		long when = (long)seconds - (long)Elapsed;
		if (when < 0) when = 0;
		return (int)when;
	}
}
