// The source code in this file was authored by
// individuals and organizations as listed below.
//
// Copyright (c) 2008  David Dindorp
//
// This source code is made available for use under
// the GPLv2 or GPLv3 license (at your discretion).
// See the file "COPYING" for details.  Contact the Free
// Software Foundation organization for more information.
// Some MySQL-specific exceptions apply, see the file "EXCEPTIONS".
// Do not remove or reduce this copyright notice.


using System.Threading;


/// <summary>
/// Helper class used to maintain clean and simple source code
/// locality given a threading implementation in C# that does
/// not exactly facilitate this (Thread is a sealed class;
/// compare Java).
///
/// Implemented as a simple wrapper around Thread
/// that executes delegated code passed to it.
///
/// Exceptions must be handled by the passed code.
/// If an unexpected exception occurs, the runtime
/// will terminate the application.
///
/// For the interrupt mechanism to work, passed code must
/// set the ready flag once it is ready to catch any
/// ThreadInterrupted exceptions.
///
/// Waiting for an OS wait handle such as a *ResetEvent does
/// not count as a WaitSleepJoin as far as Interrupt() is
/// concerned.  No InterruptedException() is thrown at a point
/// when an Interrupt() occurs while the code is waiting on
/// such an object.
/// </summary>
internal class InterruptibleThread {
	// Use this delegate when creating a worker code block.
	public delegate void WorkerCode(ManualResetEvent ready);

	// Use this property to access the internal Thread.
	private Thread wrappedThread;
	// Used to coordinate when it's appropriate to use Interrupt().
	private ManualResetEvent codeReady = new ManualResetEvent(false);
	// Contains code to be executed by this thread.
	private volatile WorkerCode code = null;

	// Used internally as the main thread method.
	private void codeWrapper() {
		code(codeReady);
	}

	/// <summary>
	/// Creates a ready-to-run thread.
	/// </summary>
	public InterruptibleThread(string threadName, int stackSize) {
		ThreadStart ts = new ThreadStart(codeWrapper);
		wrappedThread = new Thread(ts, stackSize);
		wrappedThread.IsBackground = true;
		wrappedThread.Name = threadName;
	}

	/// <summary>
	/// Schedule code to run on this thread.
	/// </summary>
	public void Run(WorkerCode code) {
		this.code = code;
		wrappedThread.Start();
	}

	/// <summary>
	/// Interrupt the next Thread.Sleep in the currently running code.
	/// </summary>
	public void Interrupt() {
		// Allow the code to exit any OS-specific WaitSleepJoin states first, then interrupt.
		codeReady.WaitOne();
		wrappedThread.Interrupt();
	}
}
