// The source code in this file was authored by
// individuals and organizations as listed below.
//
// Copyright (c) 2008  David Dindorp
//
// This source code is made available for use under
// the GPLv2 or GPLv3 license (at your discretion).
// See the file "COPYING" for details.  Contact the Free
// Software Foundation organization for more information.
// Some MySQL-specific exceptions apply, see the file "EXCEPTIONS".
// Do not remove or reduce this copyright notice.


using System;
using System.Collections.Generic;
using System.Text;


/// <summary>
/// Can parse connection strings into single key/value pairs,
/// and can reformat a single key or value into it's escaped format.
///
/// Not safe for multithreading.
/// </summary>
public class CsParseEngine {
	public delegate void AddKeyValuePairCallback(string key, string value);

	public CsParseEngine(AddKeyValuePairCallback callback) {
		this.callback = callback;
	}

	public void Parse(string input) {
		this.input = input;
		Parse();
	}

	public static string EscapeKey(string key) {
		return Escape(key);
	}

	public static string EscapeValue(string value) {
		return Escape(value);
	}

	private AddKeyValuePairCallback callback;
	private string input;
	private StringBuilder key = new StringBuilder();
	private StringBuilder value = new StringBuilder();
	private StringBuilder tailTrim = new StringBuilder();
	private int len;
	private bool inSingle, inDouble;
	private bool inValue;
	private bool keyIsNull, valueIsNull;
	private char cur, next;
	private int pos;

	private StringBuilder buffer { get { return inValue ? value : key; } }
	private bool isNull { get { return inValue ? valueIsNull : keyIsNull; } }

	private static string Escape(string input) {
		// Null strings escape to unquoted literals,
		// as per Microsoft standard practice.
		if (input == null) return string.Empty;

		// Empty strings escape to quoted literals,
		// as per Microsoft standard practice.
		if (input == string.Empty) return "\"\"";

		// Scan input string.
		bool hasSingleQuote = input.IndexOf('\'') > -1;
		bool hasDoubleQuote = input.IndexOf('"') > -1;
		bool hasSpace = input.IndexOf(' ') > -1;

		// Escape only if necessary, and in the most readable form possible.
		bool escapeWithSingle = false;
		bool escapeWithDouble = false;
		if (hasDoubleQuote && (! hasSingleQuote)) escapeWithSingle = true;
		else if (hasSingleQuote || hasSpace) escapeWithDouble = true;

		// If nothing to do, do nothing.
		if (! (escapeWithSingle || escapeWithDouble)) return input;

		// Parse and escape.
		StringBuilder output = new StringBuilder();
		if (escapeWithSingle) output.Append('\'');
		if (escapeWithDouble) output.Append('"');
		for (int i = 0; i < input.Length; i++) {
			char cur = input[i];
			if ((cur == '\'') && escapeWithSingle) output.Append("''");
			else if ((cur == '"') && escapeWithDouble) output.Append('"');
			else output.Append(cur);
		}
		if (escapeWithSingle) output.Append('\'');
		if (escapeWithDouble) output.Append('"');

		return output.ToString();
	}

	private void Parse() {
		len = input.Length;
		inSingle = false;
		inDouble = false;
		inValue = false;
		ResetBuffers();

		for (pos = 0; pos < len; pos++) {
			cur = input[pos];
			next = input.Length > pos + 1 ? input[pos + 1] : ' ';

			if (! (inSingle || inDouble)) {
				if ((cur == '=') && (! inValue) && (next != '=')) {
					// Switch from key to value marker reached.
					inValue = true;
					ResetTail();
					continue;
				}
				if (cur == ';') {
					// Next item marker reached.
					SubmitKeyValuePair();
					continue;
				}
			}

			if ((cur == '=') && (next == '=')) {
				// Escaped equals sign reached.
				AddEscapedChar();
				continue;
			}

			if ((cur == '\'') && (! inDouble) && (inSingle || isNull)) {
				if ((next == '\'') && inSingle) {
					// Escaped single quote reached.
					AddEscapedChar();
					continue;
				}
				// Single quote reached.
				UntrimTail();
				SetNotNull();
				inSingle = ! inSingle;
				continue;
			}

			if ((cur == '"') && (! inSingle) && (inDouble || isNull)) {
				if ((next == '"') && inDouble) {
					// Escaped double quote reached.
					AddEscapedChar();
					continue;
				}
				// Double quote reached.
				UntrimTail();
				SetNotNull();
				inDouble = ! inDouble;
				continue;
			}

			// Non-escape(d) character found.
			AddChar();
		}
		// End of input string reached.
		SubmitKeyValuePair();
	}

	private void ResetBuffers() {
		// Clear result buffers.
		key.Length = 0;
		value.Length = 0;
		tailTrim.Length = 0;
		keyIsNull = true;
		valueIsNull = true;
		inValue = false;
	}

	private void SubmitKeyValuePair() {
		// Submit and clear result buffers.
		if (keyIsNull && valueIsNull) {
			// Ignore the parameter string consisting of a single '='.
			// Ignore the end of the input string.
			return;
		}
		string resKey = keyIsNull ? null : key.ToString();
		string resValue = valueIsNull ? null : value.ToString();
		callback(resKey, resValue);
		ResetBuffers();
	}

	private void SetNotNull() {
		// Explicitly marks a possible empty buffer to be a non-null string.
		if (inValue) valueIsNull = false;
		else keyIsNull = false;
	}

	private void UntrimTail() {
		// Merge tail trim buffer.
		if (tailTrim.Length == 0) return;
		if (inValue) value.Append(tailTrim.ToString());
		else key.Append(tailTrim.ToString());
		SetNotNull();
		ResetTail();
	}

	private void ResetTail() {
		// Reset tail trim buffer.
		tailTrim.Length = 0;
	}

	private void AddChar() {
		if ((cur == ' ') && (! (inSingle || inDouble))) {
			if (isNull) {
				// Trim unquoted whitespace at start of value.
				return;
			} else {
				// Store unquoted whitespace at end of value for trimming.
				tailTrim.Append(cur);
			}
		} else {
			// Apply any non-tail whitespace from tail buffer.
			UntrimTail();
			// Apply current character.
			buffer.Append(cur);
			SetNotNull();
		}
	}

	private void AddEscapedChar() {
		AddChar();
		pos++;
	}
}


/// <summary>
/// Parses and contains a set of parameters given in a connection string.
///
/// A default property gives a connection string with default parameters.
///
/// The Canonicalize() method returns a canonicalized version of the
/// original string, that is a version where each parameter is given
/// it's canonical name, and entries are sorted alphabetically and
/// escaped uniformly.
///
/// Implements value comparability semantics.
/// </summary>
public abstract class ConnectionString : IEquatable<ConnectionString> {
	protected SortedDictionary<string,string> Properties;

	protected abstract string defaultParameters { get; }

	protected CsParseEngine parser;

	public ConnectionString() {
		Properties = new SortedDictionary<string,string>();
		parser = new CsParseEngine(AddKeyValuePair);
		parser.Parse(defaultParameters);
	}

	public ConnectionString(string connectionString) : this() {
		parser.Parse(connectionString);
	}

	public string this[string key] {
		get {
			return Properties[key];
		}
	}

	public static string NormalizeBoolValue(string key, string value) {
		switch (value.ToLower()) {
			case "yes": value = "True"; break;
			case "no": value = "False"; break;
			case "true": value = "True"; break;
			case "false": value = "False"; break;
			default: throw new ArgumentException(key + " is not a proper Boolean value.", key);
		}
		return value;
	}

	public static string NormalizeKey(string key, string canonical, params string[] lcAlternatives) {
		string lcKey = key.ToLower();
		foreach (string s in lcAlternatives) if (s == lcKey) {
			key = canonical;
			break;
		}
		return key;
	}

	public static string CamelCase(string input) {
		// Canonicalize to standard DSN-less CamelCase-with-spaces format.
		if (input == null) return null;
		StringBuilder output = new StringBuilder(input.Length);
		char prev = ' ';
		for (int i = 0; i < input.Length; i++) {
			if (prev == ' ') output.Append(char.ToUpper(input[i]));
			else output.Append(char.ToLower(input[i]));
			prev = input[i];
		}
		return output.ToString();
	}

	public virtual void CanonicalizeKeyValue(ref string key, ref string value) {
		key = CamelCase(key);
		// Override in subclasses to provide custom parsing.
	}

	private void AddKeyValuePair(string key, string value) {
		CanonicalizeKeyValue(ref key, ref value);
		Properties[key] = value;
	}

	public string Canonicalize() {
		// Depends on the dictionary being sorted to
		// produce a canonical representation of all
		// the properties denoting this connection string.
		StringBuilder result = new StringBuilder();
		foreach (string key in Properties.Keys) {
			string value = Properties[key];
			if (result.Length > 0) result.Append(';');
			result.Append(CsParseEngine.EscapeKey(key));
			result.Append('=');
			result.Append(CsParseEngine.EscapeValue(value));
		}
		return result.ToString();
	}

	public override int GetHashCode() {
		// The string class has en excellent GetHashCode() implementation.
		return Canonicalize().GetHashCode();
	}

	public bool Equals(ConnectionString other) {
		if (other == null) return false;
		if (other.Canonicalize() == Canonicalize()) return true;
		return false;
	}

	public override bool Equals(object obj) {
		// Yields null if the object is not castable to ConnectionString.
		ConnectionString other = obj as ConnectionString;
		// Fall back to the (better performing) generic version.
		return Equals(other);
	}

	public static bool operator ==(ConnectionString a, ConnectionString b) {
		if (object.ReferenceEquals(a, b)) return true;

		if (((object) a == null) || ((object) b == null)) return false;

		return a.Equals(b);
	}

	public static bool operator !=(ConnectionString a, ConnectionString b) {
		return ! (a == b);
	}

	public override string ToString() {
		return Canonicalize();
	}
}
