using System;
using System.Diagnostics;
using System.Threading;

namespace TestPooling {

	// Configuration class for the null driver (see further below).
	class NullDriverConfiguration : ConnectionString, IPoolConfiguration {
		// default connection string, just
		// enough to use as pooling parameters.
		protected override string defaultParameters {
			get {
				return
					"min Pool Size =0;" +
					" Max pool Size= 1000 ;" +
					"\"Connection retire\"=90;" +
					"'PoOl Retire'= \"900\" ;" +
					"Force reuse   =fAlse;" +
					"Ping Interval=10;";
			}
		}

		public NullDriverConfiguration(string connStr) : base(connStr) { }

		// Canonicalization test.
		public override void CanonicalizeKeyValue(ref string key, ref string value) {
			base.CanonicalizeKeyValue(ref key, ref value);

			// Normalize keys.
			key = NormalizeKey(key, "User Id", "uid", "user name", "username");
			key = NormalizeKey(key, "Password", "pwd");

			// Normalize values: binary.
			if (key == "Pooling" || key == "Force Reuse") {
				value = NormalizeBoolValue(key, value);
			}
		}

		// Implementation necessary for IPoolConfiguration to work.
		public int minPoolSize { get { return Convert.ToInt32(Properties["Min Pool Size"]); } }
		public int maxPoolSize { get { return Convert.ToInt32(Properties["Max Pool Size"]); } }
		public int dormantDriverRetire { get { return Convert.ToInt32(Properties["Connection Retire"]); } }
		public int dormantPoolRetire { get { return Convert.ToInt32(Properties["Pool Retire"]); } }
		public bool forceReuse { get { return Convert.ToBoolean(Properties["Force Reuse"]); } }
		public int pingInterval { get { return Convert.ToInt32(Properties["Ping Interval"]); } }
		public string poolName { get { return Properties["User Id"] + "@" + Properties["Host"]; } }
	}


	// A driver that doesn't do anything besides
	// throw an exception now and then.
	class NullDriver : IPoolableDriver {
		private static int id = 0;
		private int myId;
		private int myEvent = 0;

		public NullDriver(IPoolConfiguration config) {
			myId = ++id;
		}

		public event EventHandler QuerySuccess;

		public void OnQuerySuccess() {
			if (QuerySuccess != null) QuerySuccess();
		}

		public int ServerConnectionId {
			get { return myId; }
		}

		public bool FlushSupported {
			get { return false; }
		}

		private Random r = new Random();
		private void DieRandomly() {
			if (r.Next(50) == 17) {
				log("Simulate bad stuff on server...");
				throw new Exception("Simulate bad stuff on server");
			}
		}

		public void log(string msg) {
			Trace.WriteLine(string.Format(
				"{0:d3} {1:d3} {2} {3}",
				myId, myEvent++,
				DateTime.UtcNow.ToLongTimeString(), msg
			));
		}

		public void Flush() {
			log("Flush simulated");
			Execute();
		}

		public void Open() {
			log("Open simulated");
			DieRandomly();
		}

		public void Close() {
			log("Close simulated");
			DieRandomly();
		}

		public void Ping() {
			log("Ping simulated");
			DieRandomly();
		}

		public void Execute() {
			log("Execute simulated");
			DieRandomly();
			OnQuerySuccess();
		}
	}


	// A Connection object for the null driver.
	class NullConnection {
		// The pool manager is a singleton.
		static ConnectionPoolManager PoolManager;
		// A driver factory is needed for the pool manager.
		static IPoolableDriver factory(IPoolConfiguration config) {
			return new NullDriver((NullDriverConfiguration) config);
		}
		// A static constructor initializes the pool manager.
		static NullConnection() {
			PoolManager = new ConnectionPoolManager(factory);
		}

		private NullDriverConfiguration config;
		private NullDriver driver;
		private PooledDriverLink link;
		private ConnectionPool pool = null;

		public NullConnection(string connStr) {
			Open(connStr);
		}

		public void Open(string connStr) {
			config = new NullDriverConfiguration(connStr);
			if (config["Pooling"] == "True") {
				// If pooling enabled in config string, join a pool.
				pool = PoolManager.GetPool(config);
				// And grab a driver.
				link = pool.Catch();
				driver = (NullDriver) link.Driver;
				driver.log("Driver bound to connection");
			} else {
				// Otherwise create our own driver.
				driver = new NullDriver(config);
				driver.Open();
			}
		}

		public void Close() {
			if (pool != null) {
				// Release driver if pooling.
				if (driver != null) driver.log("Releasing driver from connection");
				pool.Release(ref link);
				if (driver != null) driver.log("Driver released from connection");
				driver = null;
			} else {
				// Disband driver otherwise.
				driver.Close();
				driver = null;
			}
		}

		public void Execute() {
			if (driver == null) throw new Exception("Not connected!");
			driver.Execute();
		}
	}
}
